import re

import sys

from sly import Parser
from sly.yacc import SlyLogger
from json5.tokenizer import JSONLexer, tokenize
from json5.model import *
from json5.utils import JSON5DecodeError
import ast
from functools import lru_cache

class QuietSlyLogger(SlyLogger):
    def warning(self, *args, **kwargs):
        return
    debug = warning
    info = warning


ESCAPE_SEQUENCES = {
    'b': '\u0008',
    'f': '\u000C',
    'n': '\u000A',
    'r': '\u000D',
    't': '\u0009',
    'v': '\u000B',
    '0': '\u0000',
    '\\': '\u005c',
    '"': '\u0022',
    "'": '\u0027',
}

# class TrailingComma:
#     pass

def replace_escape_literals(matchobj):
    s = matchobj.group(0)
    if s.startswith('\\0') and len(s) == 3:
        raise JSON5DecodeError("'\\0' MUST NOT be followed by a decimal digit", None)
    seq = matchobj.group(1)
    return ESCAPE_SEQUENCES.get(seq, seq)


@lru_cache(maxsize=1024)
def _latin_escape_replace(s):
    if len(s) != 4:
        raise JSON5DecodeError("'\\x' MUST be followed by two hexadecimal digits", None)
    val = ast.literal_eval(f'"{s}"')
    if val == '\\':
        val = '\\\\'  # this is important; the subsequent regex will sub it back to \\
    return val


def latin_escape_replace(matchobj):
    s = matchobj.group(0)
    return _latin_escape_replace(s)


class JSONParser(Parser):
    # debugfile = 'parser.out'
    tokens = JSONLexer.tokens
    log = QuietSlyLogger(sys.stderr)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.errors = []


    @_('{ wsc } value { wsc }')
    def text(self, p):
        node = JSONText(value=p[1])
        for wsc in p.wsc0:
            node.wsc_before.append(wsc)
        for wsc in p.wsc1:
            node.wsc_after.append(wsc)
        return node

    @_('key { wsc } COLON { wsc } value { wsc }')
    def first_key_value_pair(self, p):
        key = p[0]
        for wsc in p.wsc0:
            key.wsc_after.append(wsc)
        value = p[4]
        for wsc in p.wsc1:
            value.wsc_before.append(wsc)
        for wsc in p.wsc2:
            value.wsc_after.append(wsc)
        return KeyValuePair(key=p.key, value=p.value)


    @_('COMMA { wsc } [ first_key_value_pair ]')
    def subsequent_key_value_pair(self, p):
        if p.first_key_value_pair:
            node = p.first_key_value_pair
            for wsc in p.wsc:
                node.key.wsc_before.append(wsc)
        else:
            node = TrailingComma()
            for wsc in p.wsc:
                node.wsc_after.append(wsc)
        return node


    @_('WHITESPACE',
       'comment')
    def wsc(self, p):
        return p[0]

    @_('BLOCK_COMMENT')
    def comment(self, p):
        return BlockComment(p[0])


    @_('LINE_COMMENT')
    def comment(self, p):
        return LineComment(p[0])



    @_('first_key_value_pair { subsequent_key_value_pair }')
    def key_value_pairs(self, p):
        ret = [p.first_key_value_pair, ]
        for kvp in p.subsequent_key_value_pair:
            if isinstance(kvp, TrailingComma):
                return ret, kvp
            ret.append(kvp)
        return ret, None



    @_('LBRACE { wsc } [ key_value_pairs ] RBRACE')
    def json_object(self, p):
        if not p.key_value_pairs:
            node = JSONObject(leading_wsc=list(p.wsc or []))
        else:
            kvps, trailing_comma = p.key_value_pairs
            node = JSONObject(*kvps, trailing_comma=trailing_comma, leading_wsc=list(p.wsc or []))

        return node

    @_('value { wsc }')
    def first_array_value(self, p):
        node = p[0]
        for wsc in p.wsc:
            node.wsc_after.append(wsc)
        return node

    @_('COMMA { wsc } [ first_array_value ]')
    def subsequent_array_value(self, p):
        if p.first_array_value:
            node = p.first_array_value
            for wsc in p.wsc:
                node.wsc_before.append(wsc)
        else:
            node = TrailingComma()
            for wsc in p.wsc:
                node.wsc_after.append(wsc)
        return node

    @_('first_array_value { subsequent_array_value }')
    def array_values(self, p):
        ret = [p.first_array_value, ]
        for value in p.subsequent_array_value:
            if isinstance(value, TrailingComma):
                return ret, value
            ret.append(value)
        return ret, None


    @_('LBRACKET { wsc } [ array_values ] RBRACKET')
    def json_array(self, p):
        if not p.array_values:
            node = JSONArray()
        else:
            values, trailing_comma = p.array_values
            node = JSONArray(*values, trailing_comma=trailing_comma)

        for wsc in p.wsc:
            node.leading_wsc.append(wsc)

        return node

    @_('NAME')
    def identifier(self, p):
        return Identifier(name=p[0])

    @_('identifier',
       'string')
    def key(self, p):
        node = p[0]
        return node

    @_('INTEGER')
    def number(self, p):
        return Integer(p[0])

    @_('FLOAT')
    def number(self, p):
        return Float(p[0])

    @_('INFINITY')
    def number(self, p):
        return Infinity()

    @_('NAN')
    def number(self, p):
        return NaN()

    @_('MINUS number')
    def value(self, p):
        if isinstance(p.number, Infinity):
            p.number.negative = True
        node = UnaryOp(op='-', value=p.number)
        return node

    @_('PLUS number')
    def value(self, p):
        node = UnaryOp(op='+', value=p.number)
        return node

    @_('INTEGER EXPONENT',
       'FLOAT EXPONENT')
    def number(self, p):
        exp_notation = p[1][0]  # e or E
        return Float(p[0]+p[1], exp_notation=exp_notation)

    @_('HEXADECIMAL')
    def number(self, p):
        return Integer(p[0], is_hex=True)

    @_('DOUBLE_QUOTE_STRING')
    def double_quoted_string(self, p):
        raw_value = p[0]
        contents = raw_value[1:-1]
        if re.search(r'(?<!\\)([\u000D\u2028\u2029]|(?<!\r)\n)', contents):
            errmsg = f"Illegal line terminator without continuation"
            self.errors.append(JSON5DecodeError(errmsg, p._slice[0]))
        contents = re.sub(r'\\(\r\n|[\u000A\u000D\u2028\u2029])', '', contents)
        try:
            contents = re.sub(r'\\x[a-fA-F0-9]{0,2}', latin_escape_replace, contents)
        except JSON5DecodeError as exc:
            self.errors.append(JSON5DecodeError(exc.args[0], p._slice[0]))
        try:
            contents = re.sub(r'\\(0\d|.)', replace_escape_literals, contents)
        except JSON5DecodeError as exc:
            self.errors.append(JSON5DecodeError(exc.args[0], p._slice[0]))
        return DoubleQuotedString(contents, raw_value=raw_value)

    @_("SINGLE_QUOTE_STRING")
    def single_quoted_string(self, p):
        raw_value = p[0]
        contents = raw_value[1:-1]
        if re.search(r'(?<!\\)([\u000D\u2028\u2029]|(?<!\r)\n)', contents):
            errmsg = f"Illegal line terminator without continuation"
            self.errors.append(JSON5DecodeError(errmsg, p._slice[0]))
        contents = re.sub(r'\\(\r\n|[\u000A\u000D\u2028\u2029])', '', contents)
        try:
            contents = re.sub(r'\\x[a-fA-F0-9]{0,2}', latin_escape_replace, contents)
        except JSON5DecodeError as exc:
            self.errors.append(JSON5DecodeError(exc.args[0], p._slice[0]))
        try:
            contents = re.sub(r'\\(0\d|.)', replace_escape_literals, contents)
        except JSON5DecodeError as exc:
            self.errors.append(JSON5DecodeError(exc.args[0], p._slice[0]))
        return SingleQuotedString(contents, raw_value=raw_value)

    @_('double_quoted_string',
       'single_quoted_string')
    def string(self, p):
        return p[0]

    @_('TRUE')
    def boolean(self, p):
        return BooleanLiteral(True)

    @_('FALSE')
    def boolean(self, p):
        return BooleanLiteral(False)

    @_('NULL')
    def null(self, p):
        return NullLiteral()

    @_('string',
       'json_object',
       'json_array',
       'boolean',
       'null',
       'number',)
    def value(self, p):
        node = p[0]
        return node

    def error(self, token):
        if token:
            self.errors.append(JSON5DecodeError('Syntax Error', token))
        else:
            self.errors.append(JSON5DecodeError('Expecting value. Received unexpected EOF', None))

    def parse(self, *args, **kwargs):
        model = super().parse(*args, **kwargs)
        if self.errors:
            if len(self.errors) > 1:
                primary_error = self.errors[0]
                additional_errors = '\n\t'.join(err.args[0] for err in self.errors[1:])
                msg = "There were multiple errors parsing the JSON5 document.\n" \
                      "The primary error was: \n\t{}\n" \
                      "Additionally, the following errors were also detected:\n\t {}"
                msg = msg.format(primary_error.args[0], additional_errors)
                err = JSON5DecodeError(msg, None)
                err.lineno = primary_error.lineno
                err.token = primary_error.token
                err.index = primary_error.index
                raise err
            else:
                raise self.errors[0]
        return model


def parse_tokens(raw_tokens):
    parser = JSONParser()
    return parser.parse(raw_tokens)


def parse_source(text):
    tokens = tokenize(text)
    model = parse_tokens(tokens)
    return model



import math
from types import SimpleNamespace
from functools import lru_cache
from decimal import Decimal



class Node(SimpleNamespace):
    excluded_names = ['excluded_names', 'wsc_before', 'wsc_after', 'leading_wsc']
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Whitespace/Comments before/after the node
        self.wsc_before = []
        self.wsc_after = []

    def __repr__(self):
        rep = (
            f"{self.__class__.__name__}("
            + ", ".join(
                "{key}={value}".format(key=key, value=repr(value))
                for key, value in self.__dict__.items() if key not in self.excluded_names
            )
            + ")"
        )
        return rep


class JSONText(Node):
    def __init__(self, value):
        assert isinstance(value, Value)
        super().__init__(value=value)


class Value(Node):
    pass

class Key(Node):
    ...


class JSONObject(Value):
    def __init__(self, *key_value_pairs, trailing_comma=None, leading_wsc=None):
        key_value_pairs = list(key_value_pairs)
        for kvp in key_value_pairs:
            assert isinstance(kvp, KeyValuePair), f"Expected key value pair, got {type(kvp)}"
        assert leading_wsc is None or all(isinstance(item, str) or isinstance(item, Comment) for item in leading_wsc)
        super().__init__(key_value_pairs=key_value_pairs, trailing_comma=trailing_comma, leading_wsc=leading_wsc or [])


class JSONArray(Value):
    def __init__(self, *values, trailing_comma=None, leading_wsc=None):
        values = list(values)
        for value in values:
            assert isinstance(value, Value), f"Was expecting object with type Value. Got {type(value)}"
        assert leading_wsc is None or all(isinstance(item, str) or isinstance(item, Comment) for item in leading_wsc)
        super().__init__(values=values, trailing_comma=trailing_comma, leading_wsc=leading_wsc or [])


class KeyValuePair(Node):
    def __init__(self, key, value):
        assert isinstance(key, Key)
        assert isinstance(value, Value)
        super().__init__(key=key, value=value)


class Identifier(Key):
    def __init__(self, name):
        assert isinstance(name, str)
        assert len(name) > 0
        super().__init__(name=name)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return hash(self) == hash(other)


class Number(Value):
    ...


class Integer(Number):
    def __init__(self, raw_value, is_hex=False):
        value = int(raw_value) if not is_hex else int(raw_value, 0)
        super().__init__(raw_value=raw_value, value=value, is_hex=is_hex)


class Float(Number):
    def __init__(self, raw_value, exp_notation=None):
        value = float(raw_value)
        assert exp_notation is None or exp_notation in ('e', 'E')
        super().__init__(raw_value=raw_value, value=value, exp_notation=exp_notation)


class Infinity(Number):
    def __init__(self, negative=False):
        super().__init__(negative=negative)

    @property
    def value(self):
        return math.inf if not self.negative else -math.inf

    @property
    def const(self):
        return '-Infinity' if self.negative else 'Infinity'

class NaN(Number):
    def __init__(self):
        super().__init__(value=math.nan)

    @property
    def const(self):
        return 'NaN'

class String(Value, Key):
    ...

class DoubleQuotedString(String):
    def __init__(self, characters, raw_value):
        assert isinstance(raw_value, str)
        characters = characters
        assert isinstance(characters, str)
        super().__init__(characters=characters, raw_value=raw_value)


class SingleQuotedString(String):
    def __init__(self, characters, raw_value):
        assert isinstance(raw_value, str)
        characters = characters
        assert isinstance(characters, str)
        super().__init__(characters=characters, raw_value=raw_value)


class BooleanLiteral(Value):
    def __init__(self, value):
        assert value in (True, False)
        super().__init__(value=value)


class NullLiteral(Value):
    def __init__(self):
        super().__init__(value=None)


class UnaryOp(Value):
    def __init__(self, op, value):
        assert op in ('-', '+')
        assert isinstance(value, Number)
        super().__init__(op=op, value=value)

class TrailingComma(Node):
    ...


class Comment(Node):
    def __init__(self, value):
        assert isinstance(value, str), f"Expected str got {type(value)}"
        super().__init__(value=value)


class LineComment(Comment):
    ...

class BlockComment(Comment):
    ...



import re
import sys
from sly import Lexer
from sly.lex import Token
from json5.utils import JSON5DecodeError
import logging

logger = logging.getLogger(__name__)
# logger.addHandler(logging.StreamHandler(stream=sys.stderr))
# logger.setLevel(level=logging.DEBUG)
class JSON5Token(Token):
    '''
    Representation of a single token.
    '''
    def __init__(self, tok, doc):
        self.type = tok.type
        self.value = tok.value
        self.lineno = tok.lineno
        self.index = tok.index
        self.doc = doc
    __slots__ = ('type', 'value', 'lineno', 'index', 'doc')

    def __repr__(self):
        return f'JSON5Token(type={self.type!r}, value={self.value!r}, lineno={self.lineno}, index={self.index})'


class JSONLexer(Lexer):
    reflags = re.DOTALL
    tokens = {LBRACE, RBRACE,
              LBRACKET, RBRACKET,
              DOUBLE_QUOTE_STRING, SINGLE_QUOTE_STRING,
              NAME,
              COMMA,
              BLOCK_COMMENT,
              LINE_COMMENT,
              WHITESPACE,
              TRUE, FALSE, NULL,
              COLON,

              # Numbers
              PLUS, MINUS,
              FLOAT, INTEGER,
              INFINITY, NAN, EXPONENT,
              HEXADECIMAL
              }

    def tokenize(self, text, *args, **kwargs):
        for tok in super().tokenize(text, *args, **kwargs):
            tok = JSON5Token(tok, text)
            yield tok

    LBRACE = r'{'
    RBRACE = r'}'
    LBRACKET = r'\['
    RBRACKET = r'\]'
    COLON = r"\:"
    COMMA = r"\,"

    DOUBLE_QUOTE_STRING = r'"(?:[^"\\]|\\.)*"'
    SINGLE_QUOTE_STRING = r"'(?:[^'\\]|\\.)*'"

    LINE_COMMENT = r"//[^\n]*"
    @_(r'/\*((.|\n))*?\*/')
    def BLOCK_COMMENT(self, tok):
        self.lineno += tok.value.count('\n')
        return tok

    @_("[\u0009\u000A\u000B\u000C\u000D\u0020\u00A0\u2028\u2029\ufeff]+")
    def WHITESPACE(self, tok):
        self.lineno += tok.value.count('\n')
        return tok

    MINUS = r'\-'
    PLUS = r'\+'
    EXPONENT = r"(e|E)(\-|\+)?\d+"
    HEXADECIMAL = r'0(x|X)[0-9a-fA-F]+'
    FLOAT = r'(\d+\.\d*)|(\d*\.\d+)'      # 23.45
    INTEGER = r'\d+'
    NAME = r'[a-zA-Z_\$]([a-zA-Z_\d\$])*'

    NAME['true'] = TRUE
    NAME['false'] = FALSE
    NAME['null'] = NULL
    NAME['Infinity'] = INFINITY
    NAME['NaN'] = NAN

    def error(self, t):
        raise JSON5DecodeError(f'Illegal character {t.value[0]!r} at index {self.index}', None)

def tokenize(text):
    lexer = JSONLexer()
    tokens = lexer.tokenize(text)
    return tokens



import types

import sys

from json5.parser import parse_source
from json5.model import *
from json5.utils import singledispatchmethod
from collections import UserString


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(level=logging.DEBUG)
# logger.addHandler(logging.StreamHandler(stream=sys.stderr))

class Environment(types.SimpleNamespace):
    def __init__(self, object_hook=None, parse_float=None, parse_int=None, parse_constant=None, strict=True, object_pairs_hook=None):
        super().__init__(object_hook=object_hook, parse_float=parse_float, parse_int=parse_int, parse_constant=parse_constant, strict=strict, object_pairs_hook=object_pairs_hook)


class JsonIdentifier(UserString):
    ...

def load(f, **kwargs):
    """
    Like loads, but takes a file-like object with a read method.

    :param f:
    :param kwargs:
    :return:
    """
    text = f.read()
    return loads(text)

def loads(s, *args, loader=None, **kwargs):
    """
    Take a string of JSON text and deserialize it

    :param s:
    :param loader: The loader class to use
    :param object_hook: same meaning as in ``json.loads``
    :param parse_float: same meaning as in ``json.loads``
    :param parse_int: same meaning as in ``json.loads``
    :param parse_constant: same meaning as in ``json.loads``
    :param strict: same meaning as in ``json.loads`` (currently has no effect)
    :param object_pairs_hook: same meaning as in ``json.loads``
    :return:
    """
    model = parse_source(s)
    # logger.debug('Model is %r', model)
    if loader is None:
        loader = DefaultLoader(**kwargs)
    return loader.load(model)

class DefaultLoader:
    def __init__(self, env=None, **env_kwargs):
        if env is None:
            env = Environment(**env_kwargs)
        self.env = env

    @singledispatchmethod
    def load(self, node):
        raise NotImplementedError(f"Can't load node {node}")

    to_python = load.register

    @to_python(JSONText)
    def json_model_to_python(self, node):
        logger.debug('json_model_to_python evaluating node %r', node)
        return self.load(node.value)

    @to_python(JSONObject)
    def json_object_to_python(self, node):
        logger.debug('json_object_to_python evaluating node %r', node)
        d = {}
        for key_value_pair in node.key_value_pairs:
            key = self.load(key_value_pair.key)
            value = self.load(key_value_pair.value)
            d[key] = value
        if self.env.object_pairs_hook:
            return self.env.object_pairs_hook(list(d.items()))
        elif self.env.object_hook:
            return self.env.object_hook(d)
        else:
            return d


    @to_python(JSONArray)
    def json_array_to_python(self, node):
        logger.debug('json_array_to_python evaluating node %r', node)
        return [self.load(value) for value in node.values]

    @to_python(Identifier)
    def identifier_to_python(self, node):
        logger.debug('identifier_to_python evaluating node %r', node)
        return JsonIdentifier(node.name)


    @to_python(Number)  # NaN/Infinity are covered here
    def number_to_python(self, node):
        logger.debug('number_to_python evaluating node %r', node)
        if self.env.parse_constant:
            return self.env.parse_constant(node.const)
        return node.value

    @to_python(Integer)
    def integer_to_python(self, node):
        if self.env.parse_int:
            return self.env.parse_int(node.raw_value)
        else:
            return node.value

    @to_python(Float)
    def float_to_python(self, node):
        if self.env.parse_float:
            return self.env.parse_float(node.raw_value)
        else:
            return node.value

    @to_python(UnaryOp)
    def unary_to_python(self, node):
        logger.debug('unary_to_python evaluating node %r', node)
        if isinstance(node.value, Infinity):
            return self.load(node.value)
        value = self.load(node.value)
        if node.op == '-':
            return value * -1
        else:
            return value

    @to_python(String)
    def string_to_python(self, node):
        logger.debug('string_to_python evaluating node %r', node)
        return node.characters


    @to_python(NullLiteral)
    def null_to_python(self, node):
        logger.debug('null_to_python evaluating node %r', node)
        return None

    @to_python(BooleanLiteral)
    def boolean_to_python(self, node):
        logger.debug('boolean_to_python evaluating node %r', node)
        return node.value

    @to_python(Comment)
    def comment_or_whitespace_to_python(self, node):
        raise RuntimeError("Comments are not supported in the default loader!")


class ModelLoader(DefaultLoader):
    def load(self, node):
        return node



from .utils import singledispatchmethod
from json5.model import *
from collections import UserDict
import json
import io

class Environment(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.outfile = io.StringIO()
        self.indent_level = 0
        self.indent = 0


    def write(self, s, indent=None):
        if indent is None:
            indent = self.indent_level
        whitespace = ' ' * self.indent * indent
        s = f'{whitespace}{s}'
        self.outfile.write(s)


def dump(obj, f, **kwargs):
    text = dumps(obj, **kwargs)
    return f.write(text)


def dumps(obj, dumper=None, indent=0):
    env = Environment()
    env.indent = indent
    if dumper is None:
        dumper = DefaultDumper(env=env)
    model = dumper.dump(obj)
    dumper.env.outfile.seek(0)
    return dumper.env.outfile.read()

class DefaultDumper:
    """
    Dump Python objects to a JSON string
    """
    def __init__(self, env=None):
        if env is None:
            env = Environment()
        self.env = env

    @singledispatchmethod
    def dump(self, obj):
        raise NotImplementedError(f"Cannot dump node {repr(obj)}")
    
    to_json = dump.register
    
    @to_json(dict)
    def dict_to_json(self, d):
        self.env.write('{', indent=0)
        if self.env.indent:
            self.env.write('\n')
            self.env.indent_level += 1
        for index, (key, value) in enumerate(d.items(), start=1):
            if self.env.indent:
                self.env.write('')
            self.dump(key)
            self.env.write(': ', indent=0)
            self.dump(value)
            if index == len(d):
                break
            if self.env.indent:
                self.env.write(',', indent=0)
                self.env.write('\n', indent=0)
            else:
                self.env.write(', ', indent=0)
    
        if self.env.indent:
            self.env.indent_level -= 1
            self.env.write('\n')
            self.env.write('}')
        else:
            self.env.write('}', indent=0)
    
    
    @to_json(int)
    def int_to_json(self, i):
        self.env.write(str(i), indent=0)
    
    
    @to_json(str)
    def str_to_json(self, s):
        self.env.write(json.dumps(s), indent=0)
        # if isinstance(s, JsonIdentifier):
        #     quote_char = None
        # elif "'" in s:
        #     quote_char = '"'
        # else:
        #     quote_char = "'"
        #
        # if quote_char:
        #     self.env.write(quote_char, indent=0)
        # self.env.write(s, indent=0)
        # if quote_char:
        #     self.env.write(quote_char, indent=0)
    
    
    @to_json(list)
    def list_to_json(self, l):
        self.env.write('[', indent=0)
        if self.env.indent:
            self.env.indent_level += 1
            self.env.write('\n', indent=0)
        list_length = len(l)
        for index, item in enumerate(l, start=1):
            if self.env.indent:
                self.env.write('')
            self.dump(item)
            if index != list_length:
                if self.env.indent:
                    self.env.write(',', indent=0)
                else:
                    self.env.write(', ', indent=0)
            if self.env.indent:
                self.env.write('\n', indent=0)
        if self.env.indent:
            self.env.indent_level -= 1
        self.env.write(']')
    
    
    @to_json(float)
    def float_to_json(self, f):
        if f == math.inf:
            self.env.write('Infinity', indent=0)
        elif f == -math.inf:
            self.env.write('-Infinity', indent=0)
        elif f is math.nan:
            self.env.write('NaN', indent=0)
        else:
            self.env.write(str(f), indent=0)

    @to_json(bool)
    def bool_to_json(self, b):
        self.env.write(str(b).lower(), indent=0)

    @to_json(type(None))
    def none_to_json(self, _):
        self.env.write('null', indent=0)

class ModelDumper:
    """
    Dump a model to a JSON string
    """
    def __init__(self, env=None):
        #  any provided environment is ignored
        self.env = Environment()

    def process_wsc_before(self, node):
        for wsc in node.wsc_before:
            if isinstance(wsc, Comment):
                self.dump(wsc)
            elif isinstance(wsc, str):
                self.env.write(wsc)
            else:
                raise ValueError(f"Did not expect {type(node)}")

    def process_wsc_after(self, node):
        for wsc in node.wsc_after:
            if isinstance(wsc, Comment):
                self.dump(wsc)
            elif isinstance(wsc, str):
                self.env.write(wsc)
            else:
                raise ValueError(f"Did not expect {type(wsc)}")

    def process_leading_wsc(self, node):
        for wsc in node.leading_wsc:
            if isinstance(wsc, Comment):
                self.dump(wsc)
            elif isinstance(wsc, str):
                self.env.write(wsc)
            else:
                raise ValueError(f"Did not expect {type(wsc)}")

    @singledispatchmethod
    def dump(self, node):
        raise NotImplementedError('foo')

    to_json = dump.register

    @to_json(JSONText)
    def json_model_to_json(self, node):
        self.process_wsc_before(node)
        self.dump(node.value)
        self.process_wsc_after(node)

    @to_json(JSONObject)
    def json_object_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write('{')
        if node.leading_wsc:
            self.process_leading_wsc(node)
        num_pairs = len(node.key_value_pairs)
        for index, kvp in enumerate(node.key_value_pairs, start=1):
            self.dump(kvp.key)
            self.env.write(':')
            self.dump(kvp.value)
            if index != num_pairs:
                self.env.write(',')
        if node.trailing_comma:
            self.dump(node.trailing_comma)
        self.env.write('}')
        self.process_wsc_after(node)

    @to_json(JSONArray)
    def json_array_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write('[')
        if node.leading_wsc:
            self.process_leading_wsc(node)
        for index, value in enumerate(node.values, start=1):
            self.dump(value)
            if index != len(node.values):
                self.env.write(',')
        if node.trailing_comma:
            self.dump(node.trailing_comma)
        self.env.write(']')
        self.process_wsc_after(node)

    @to_json(Identifier)
    def identifier_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(node.name)
        self.process_wsc_after(node)

    @to_json(Integer)
    def integer_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(node.raw_value)
        self.process_wsc_after(node)

    @to_json(Float)
    def float_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(node.raw_value)
        self.process_wsc_after(node)

    @to_json(UnaryOp)
    def unary_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(node.op)
        self.dump(node.value)
        self.process_wsc_after(node)

    @to_json(String)
    def string_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(node.raw_value)  # The original value, including any escape sequences or line continuations
        self.process_wsc_after(node)

    @to_json(NullLiteral)
    def null_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write('null')
        self.process_wsc_after(node)

    @to_json(BooleanLiteral)
    def boolean_to_json(self, node):
        self.process_wsc_before(node)
        if node.value:
            self.env.write('true')
        else:
            self.env.write('false')
        self.process_wsc_after(node)

    @to_json(LineComment)
    def line_comment_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(node.value)
        self.process_wsc_after(node)

    @to_json(BlockComment)
    def block_comment_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(node.value)
        self.process_wsc_after(node)

    @to_json(TrailingComma)
    def trailing_comma_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write(',')
        self.process_wsc_after(node)

    @to_json(Infinity)
    def infinity_to_json(self, node):
        self.process_wsc_before(node)

        self.env.write('Infinity')
        self.process_wsc_after(node)

    @to_json(NaN)
    def nan_to_json(self, node):
        self.process_wsc_before(node)
        self.env.write('NaN')
        self.process_wsc_after(node)

class Modelizer:
    """
    Turn Python objects into a model
    """
    @singledispatchmethod
    def modelize(self, obj):
        raise NotImplementedError(f"Cannot modelize object of type {type(obj)}")
    
    to_model = modelize.register
    
    @to_model(str)
    def str_to_model(self, s):
        if repr(s).startswith("'"):
            return SingleQuotedString(s, raw_value=repr(s))
        else:
            return DoubleQuotedString(s, raw_value=repr(s))
    
    @to_model(dict)
    def dict_to_model(self, d):
        kvps = []
        for key, value in d.items():
            kvp = KeyValuePair(key=self.modelize(key), value=self.modelize(value))
            kvps.append(kvp)
        return JSONObject(*kvps)
    
    @to_model(list)
    def list_to_model(self, lst):
        list_values = []
        for v in lst:
            list_values.append(self.modelize(v))
        return JSONArray(*list_values)
    
    @to_model(int)
    def int_to_model(self, i):
        return Integer(str(i))


    @to_model(float)
    def float_to_model(self, f):
        if f == math.inf:
            return Infinity()
        elif f == -math.inf:
            return UnaryOp('-', Infinity())
        elif f is math.nan:
            return NaN()
        else:
            return Float(str(f))

    @to_model(bool)
    def bool_to_model(self, b):
        return BooleanLiteral(b)

    @to_model(type(None))
    def none_to_model(self, _):
        return NullLiteral()


def modelize(obj):
    """

    :param obj: a python object
    :return: a model representing the python object
    """
    return Modelizer().modelize(obj)



from setuptools import setup
from io import open

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='json-five',
    version='0.3.5',
    packages=['json5'],
    url='https://github.com/spyoungtech/json-five',
    license='Apache',
    author='Spencer Phillip Young',
    author_email='spencer.young@spyoung.com',
    description='A JSON5 parser that, among other features, supports round-trip preservation of comments',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['sly'],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)


from functools import singledispatch, update_wrapper
from json import JSONDecodeError
try:
    from functools import singledispatchmethod
except ImportError:
    def singledispatchmethod(func):
        dispatcher = singledispatch(func)

        def wrapper(*args, **kwargs):
            return dispatcher.dispatch(args[1].__class__)(*args, **kwargs)

        wrapper.register = dispatcher.register
        update_wrapper(wrapper, func)
        return wrapper


class JSON5DecodeError(JSONDecodeError):
    def __init__(self, msg, token):
        lineno = getattr(token, 'lineno', 0)
        index = getattr(token, 'index', 0)
        doc = getattr(token, 'doc', None)
        self.token = token
        self.index = index
        if token and doc:
            errmsg = f'{msg} in or near token {token.type} at'
            super().__init__(errmsg, doc, index)
        else:
            ValueError.__init__(self, msg)
            self.msg = msg
            self.lineno = lineno

    def __reduce__(self):
        return self.__class__, (self.msg, self.token)


